@@ -4,6 +4,8 @@ import torch.nn.functional as F
 from transformers import AutoConfig, AutoModelForCausalLM, PretrainedConfig, PreTrainedModel
 from transformers.modeling_outputs import CausalLMOutput
 
+import warnings
+
 
 
 class MultiheadLeap(nn.Module):
@@ -34,11 +36,12 @@ class MultiheadLeap(nn.Module):
         k = k.reshape(batch_size, seq_len, self.n_head, self.head_size)
         v = v.reshape(batch_size, seq_len, self.n_head, self.head_size)
         
-        # normalize vectors so dot products don't get too big
+        # unparameterized norming of vectors so dot products don't explode (also why it is after reshaping)
         if self.rescale:
             q = self.__real_norm(q)
             f = self.__real_norm(f)
             k = self.__real_norm(k)
+            v = self.__real_norm(v)
             
         # dropout regularization (keys don't need dropout as they are always dotted with a dropped out vector)
         q = self.drop(q)
@@ -106,7 +109,7 @@ class LeapBlock(nn.Module):
     def __init__(self, config, window_size):
         super(LeapBlock, self).__init__()
 
-        self.pre_norm = nn.LayerNorm(config.hidden_size)
+        self.attn_norm = nn.LayerNorm(config.hidden_size)
         
         # modules for leap
         # note: one large projection matrix is equivalent to having seperate projection matrices and is faster
@@ -115,26 +118,22 @@ class LeapBlock(nn.Module):
                                   rescale = config.rescale, dropout = config.hidden_dropout_prob)
 
         # modules for feedforward layer (aka boom layer)
-        self.boom = nn.Linear(config.hidden_size, config.hidden_size * 4, bias = False)
+        self.boom = nn.Linear(config.hidden_size, config.hidden_size * 4)
         self.activation = nn.GELU()
-        self.unboom = nn.Linear(4 * config.hidden_size, config.hidden_size, bias = False)
+        self.unboom = nn.Linear(4 * config.hidden_size, config.hidden_size)
         self.boom_norm = nn.LayerNorm(config.hidden_size)
         self.boom_drop = nn.Dropout(config.hidden_dropout_prob)
 
 
     def forward(self, mod, attention_mask):
-        # pre-norming
-        mod_normed = self.pre_norm(mod)
-        
-        # projections so each matrix has its own purpose
-        q, f, k, v = self.projections(mod_normed).chunk(4, dim = -1)
+        # pre-norming with projections so each matrix has its own purpose
+        q, f, k, v = self.projections(self.attn_norm(mod)).chunk(4, dim = -1)
         
         # unnormed residual connection
         mod = mod + self.leap(q, f, k, v, attention_mask)
         
         # feedforward layer with pre-norming
-        mod_normed = self.pre_norm(mod)
-        mod = mod + self.__boom(self.boom_norm(mod_normed))
+        mod = mod + self.__boom(self.boom_norm(mod))
         
         return mod
 
@@ -197,6 +196,12 @@ class LeapConfig(PretrainedConfig):
         # check head sizes
         assert hidden_size % n_head == 0, "hidden_size is not divisible by n_head"
         
+        if (hidden_size // n_head) > 64:
+            warnings.warn("Using a hidden_size-to-head ratio of greater than 64 is not ideal as"
+                          " LEAP uses a simplified form of attention that relies on having many heads")
+        if (hidden_size // n_head) < 32:
+            warnings.warn("Using a hidden_size-to-head ratio of less than 64 can sometimes lead to instability")
+        
         # check window sizes (and set them automatically if not set)
         assert not (use_local_att is False and window_sizes is not None), \
             "window sizes set when not using local attention"
@@ -231,11 +236,11 @@ class LeapConfig(PretrainedConfig):
 class LeapForCausalLM(PreTrainedModel):
     config_class = LeapConfig
     
-    def __init__(self,config):
+    def __init__(self, config):
         super().__init__(config)
         self.config = config
         self.word_embedding = nn.Embedding(config.vocab_size, config.hidden_size)
-        self.proj_logits = nn.Linear(config.hidden_size, config.vocab_size)
+        self.proj_logits = nn.Linear(config.hidden_size, config.vocab_size, bias = False)
         self.leap_model = LeapDecoder(config)
         self.last_norm = nn.LayerNorm(config.hidden_size)
         self.criterion = nn.CrossEntropyLoss()
